@@ -7,14 +7,17 @@ use crate::types::{Key, KeyValue};
 use crate::utils::as_i32;
 use crate::Pyxel;
 
+#[derive(PartialEq)]
 enum KeyState {
-    Pressed { frame_count: u32 },
-    Released { frame_count: u32 },
+    Pressed,
+    Released,
+    PressedAndReleased,
+    ReleasedAndPressed,
 }
 
 pub struct Input {
     is_mouse_visible: bool,
-    key_states: HashMap<Key, KeyState>,
+    key_states: HashMap<Key, (u32, KeyState)>,
     key_values: HashMap<Key, KeyValue>,
     input_keys: Vec<Key>,
     input_text: String,
@@ -62,18 +65,9 @@ impl Input {
             // Key events
             Event::KeyDown { keycode } => {
                 self.press_key(keycode, frame_count);
-                if let Some(keycode) = to_integrated_key(keycode) {
-                    self.press_key(keycode, frame_count);
-                }
-                if is_keyboard_key(keycode) {
-                    self.input_keys.push(keycode);
-                }
             }
             Event::KeyUp { keycode } => {
                 self.release_key(keycode, frame_count);
-                if let Some(keycode) = to_integrated_key(keycode) {
-                    self.release_key(keycode, frame_count);
-                }
             }
             Event::TextInput { text } => {
                 self.input_text += &text;
@@ -131,13 +125,32 @@ impl Input {
     }
 
     fn press_key(&mut self, key: Key, frame_count: u32) {
-        self.key_states
-            .insert(key, KeyState::Pressed { frame_count });
+        let mut key_state = KeyState::Pressed;
+        if let Some((last_frame_count, last_key_state)) = self.key_states.get(&key) {
+            if *last_frame_count == frame_count && *last_key_state != KeyState::Pressed {
+                key_state = KeyState::ReleasedAndPressed;
+            }
+        }
+        self.key_states.insert(key, (frame_count, key_state));
+        if is_keyboard_key(key) {
+            self.input_keys.push(key);
+        }
+        if let Some(key) = to_integrated_key(key) {
+            self.press_key(key, frame_count)
+        }
     }
 
     fn release_key(&mut self, key: Key, frame_count: u32) {
-        self.key_states
-            .insert(key, KeyState::Released { frame_count });
+        let mut key_state = KeyState::Released;
+        if let Some((last_frame_count, last_key_state)) = self.key_states.get(&key) {
+            if *last_frame_count == frame_count && *last_key_state != KeyState::Released {
+                key_state = KeyState::PressedAndReleased;
+            }
+        }
+        self.key_states.insert(key, (frame_count, key_state));
+        if let Some(key) = to_integrated_key(key) {
+            self.release_key(key, frame_count)
+        }
     }
 }
 
@@ -167,10 +180,15 @@ impl Pyxel {
     }
 
     pub fn btn(&self, key: Key) -> bool {
-        matches!(
-            self.input.key_states.get(&key),
-            Some(KeyState::Pressed { .. })
-        )
+        if let Some((frame_count, key_state)) = self.input.key_states.get(&key) {
+            if *key_state == KeyState::Pressed
+                || *key_state == KeyState::ReleasedAndPressed
+                || *frame_count == self.frame_count() && *key_state == KeyState::PressedAndReleased
+            {
+                return true;
+            }
+        }
+        false
     }
 
     pub fn btnp(
@@ -179,10 +197,16 @@ impl Pyxel {
         hold_frame_count: Option<u32>,
         period_frame_count: Option<u32>,
     ) -> bool {
-        if let Some(KeyState::Pressed { frame_count }) = self.input.key_states.get(&key) {
+        if let Some((frame_count, key_state)) = self.input.key_states.get(&key) {
+            if *key_state == KeyState::Released {
+                return false;
+            }
             if *frame_count == self.frame_count() {
                 return true;
             }
+            if *key_state == KeyState::PressedAndReleased {
+                return false;
+            }
             let hold_frame_count = hold_frame_count.unwrap_or(0);
             let period_frame_count = period_frame_count.unwrap_or(0);
             if hold_frame_count == 0 || period_frame_count == 0 {
@@ -198,7 +222,10 @@ impl Pyxel {
     }
 
     pub fn btnr(&self, key: Key) -> bool {
-        if let Some(KeyState::Released { frame_count }) = self.input.key_states.get(&key) {
+        if let Some((frame_count, key_state)) = self.input.key_states.get(&key) {
+            if *key_state == KeyState::Pressed {
+                return false;
+            }
             if *frame_count == self.frame_count() {
                 return true;
             }
@@ -217,17 +244,8 @@ impl Pyxel {
     pub fn set_btn(&mut self, key: Key, key_state: bool) {
         if key_state {
             self.input.press_key(key, self.frame_count());
-            if let Some(key) = to_integrated_key(key) {
-                self.input.press_key(key, self.frame_count());
-            }
-            if is_keyboard_key(key) {
-                self.input.input_keys.push(key);
-            }
         } else {
             self.input.release_key(key, self.frame_count());
-            if let Some(key) = to_integrated_key(key) {
-                self.input.release_key(key, self.frame_count());
-            }
         }
     }
 
